/**
 * @file AECommProject.cpp
 * @brief Legt die Initialisierungsroutinen fuer die DLL fest.
 */
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "AECommProject.h"
#include "AEComm.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//
//	Beachten Sie!
//
//		Wird diese DLL dynamisch an die MFC-DLLs gebunden,
//		muss bei allen von dieser DLL exportierten Funktionen,
//		die MFC-Aufrufe durchfhren, das Makro AFX_MANAGE_STATE
//		direkt am Beginn der Funktion eingefgt sein.
//
//		Beispiel:
//
//		extern "C" BOOL PASCAL EXPORT ExportedFunction()
//		{
//			AFX_MANAGE_STATE(AfxGetStaticModuleState());
//		// Hier normaler Funktionsrumpf
//		}
//
//		Es ist sehr wichtig, dass dieses Makro in jeder Funktion
//		vor allen MFC-Aufrufen erscheint. Dies bedeutet, dass es
//		als erste Anweisung innerhalb der Funktion ausgefhrt werden
//		muss, sogar vor jeglichen Deklarationen von Objektvariablen,
//		da ihre Konstruktoren Aufrufe in die MFC-DLL generieren
//		knnten.
//
//		Siehe MFC Technical Notes 33 und 58 fr weitere
//		Details.
//

/////////////////////////////////////////////////////////////////////////////
// CAECommProjectApp

#ifdef _WIN32
BEGIN_MESSAGE_MAP(CAECommProjectApp, CWinApp)
	//{{AFX_MSG_MAP(CAECommProjectApp)
		// HINWEIS - Hier werden Mapping-Makros vom Klassen-Assistenten eingefgt und entfernt.
		//    Innerhalb dieser generierten Quelltextabschnitte NICHTS VERNDERN!
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()
#endif

/////////////////////////////////////////////////////////////////////////////
// CAECommProjectApp Konstruktion
CAECommProjectApp *CAECommProjectApp::m_this = NULL;

CAECommProjectApp::CAECommProjectApp()
{
	// ZU ERLEDIGEN: Hier Code zur Konstruktion einfgen
	// Alle wichtigen Initialisierungen in InitInstance platzieren
	m_this = this;
#ifdef WIN32
	// Members Initialisieren

	// Haupttask 
	m_workerThread[0] = NULL;
	m_workerThread[1] = NULL;

	// Sockets initialisieren
    m_listenSocket = INVALID_SOCKET;
	// Portnummer fuer Listensocket
	m_listenPort = 8000;
#endif
}

/////////////////////////////////////////////////////////////////////////////
// CAECommProjectApp Members
#ifdef _WIN32
/**
 * @brief Initialisieren der Anwendung.
 */
BOOL CAECommProjectApp::InitInstance() 
{
	// Benutzung der Socket - Funktionalitaet initialisieren
	WSADATA WindowsData;
	WORD    Version;

	Version = MAKEWORD(2, 2);
	WSAStartup(Version, &WindowsData);

#ifdef _DEBUG
//	StartListener();
//	StartNotify();
#endif

	return CWinApp::InitInstance();
}
#endif

#ifdef _WIN32
/**
 * @brief Anwendung auf beenden vorbereiten.
 */
int CAECommProjectApp::ExitInstance() 
{
#ifdef _DEBUG
	EndNotify();
	EndListener();
#endif

	// Socket - Funktionalitaet exit
	WSACleanup();

	return CWinApp::ExitInstance();
}
#endif

/**
 * @brief Sicherstellen das die Anfrage und das Kanalobjekt in Ordnung ist.
 */
bool CAECommProjectApp::AEChannelRequest(short channelID)
{
	// Kanalbereich pruefen und evntl. erstellen des nicht vorhandenen Kanalobjektes.
	if ((channelID < 0) || (channelID >= MAX_CHANNEL)) {
		
		// Bereich falsch
		return false;
	}

	// Kanalobjekt bereits vorhanden?
	if (m_channelArray[channelID] != NULL) {
	
		return true;
	}

	// Kanalobjekt erzeugen
	m_channelArray[channelID] = new CAEChannel((unsigned char)channelID);

	return true;
}

/**
 * @brief Erhalt eines Kanalobjekts oder NULL bei Bereichsueberschreitung sowie Fehler.
 */
CAEChannel *CAECommProjectApp::GetAEChannel(short channelID)
{
	if (AEChannelRequest(channelID)) {
		
		return m_channelArray[channelID];
	}

	return NULL;
}

#ifdef WIN32
/**
 * @brief Sicherstellen das die Anfrage und das Kanalobjekt in Ordnung ist.
 */
bool CAECommProjectApp::Channel4Request(short &channelID)
{
	// Kanalbereich pruefen und evntl. erstellen des nicht vorhandenen Kanalobjektes.
	if ((channelID != -1) && (channelID < 0 || channelID >= MAX_CHANNEL4)) {
		
		// Bereich falsch
		return false;
	}

	if (channelID == -1) {
		// Freies suchen
		for (channelID=0; channelID < MAX_CHANNEL4; channelID++) {

			if  (m_channel4Array[channelID] == NULL) {

				break;
			} else if (!m_channel4Array[channelID]->IsBusy()) {

				return true;
			}
		}

		if (channelID == MAX_CHANNEL4) {
		
			// Alle Kanalklassen sind vergeben!
			channelID = -1;
			return false;
		}
	} else {

		// Kanalobjekt bereits vorhanden?
		if (m_channel4Array[channelID] != NULL) {
		
			return true;
		}
	}


	// Kanalobjekt erzeugen
	m_channel4Array[channelID] = new CChannel4(channelID, IV_DEFAULT_DEVICEID);

	return true;
}

/**
 * @brief Erhalt eines Kanalobjekts oder NULL bei Bereichsueberschreitung sowie Fehler.
 */
CChannel4 *CAECommProjectApp::GetChannel4(short &channelID)
{
	if (Channel4Request(channelID)) {
		
		// CallBack fuer Notify uebergeben
		m_channel4Array[channelID]->SetCBNotifyRecord(CBNotifyRecord);
		return m_channel4Array[channelID];
	}

	return NULL;
}

int CAECommProjectApp::SetNotifyDevice(int type, char *ip, int port, int maxConnections, int state)
{
	switch(type) {
	case IV_VALUE_NOTIFY_RECORD:
		if (state) {
			if (!StartListener()) return IV_ERROR_FAIL;
		} else {
			if (!EndListener()) return IV_ERROR_FAIL;
		}
		break;
	default:
		return IV_ERROR_FAIL;
	}

	return IV_ERROR_SUCCESS;
}

int CAECommProjectApp::SetNotifySoftware(int type, char *ip, int port, int retry, int delay, int state)
{
	switch(type) {
	case IV_VALUE_NOTIFY_RECORD:
		if (state) {
			if (!StartNotify()) return IV_ERROR_FAIL;
		} else {
			if (!EndNotify()) return IV_ERROR_FAIL;
		}
		break;
	default:
		return IV_ERROR_FAIL;
	}

	return IV_ERROR_SUCCESS;
}

int CAECommProjectApp::StartListener()
{
	// Hauptthread starten; Dieser besorgt die Verbindunsannahme.
	if (m_workerThread[0] == NULL) {
	
		DWORD threadID;
		m_workerThread[0] = ::CreateThread(NULL, 0, ListenThread, this, 0, &threadID);
	}

	return (m_workerThread[0] != NULL);
}

int CAECommProjectApp::EndListener()
{
	// Hauptthread deaktivieren, wenn aktiv
	if (m_workerThread[0] != NULL) {

		// WorkerThread fuer Socket beenden.
		m_endSignal[0] = TRUE;
		
		// Thread umschalten
		Sleep(10);
		
		// Thread Beendigung abwarten (ListenThread stellt m_endSignal auf FALSE)
		int i;
		for (i=0; (i < 100) && m_endSignal[0]; i++) {
				
			Sleep(50);
		}
		
		CloseHandle(m_workerThread[0]);
		m_workerThread[0] = NULL;

		TRACE("Listener end by %i\n", i);
	}

	return !m_endSignal[0];
}

int CAECommProjectApp::StartNotify()
{
	// Benachrichtigungsthread starten; Dieser besorgt die Nachrichtenverteilung.
	if (m_workerThread[1] == NULL) {
	
		DWORD threadID;
		m_workerThread[1] = ::CreateThread(NULL, 0, NotifyThread, this, 0, &threadID);
	}

	return (m_workerThread[1] != NULL);
}

int CAECommProjectApp::EndNotify()
{
	// Benachrichtigungsthread deaktivieren, wenn aktiv
	if (m_workerThread[1] != NULL) {

		// WorkerThread fuer Benachrichtigung beenden.
		m_endSignal[1] = TRUE;
		
		// Thread umschalten
		Sleep(10);
		
		// Thread Beendigung abwarten
		int i;
		for (i=0; (i < 100) && m_endSignal[1]; i++) {
				
			Sleep(50);
		}
		
		CloseHandle(m_workerThread[1]);
		m_workerThread[1] = NULL;

		TRACE("Notifyer end by %i\n", i);
	}

	return !m_endSignal[1];
}

unsigned long WINAPI CAECommProjectApp::ListenThread(LPVOID lpParam)
{
	// Muss einen Parameter uebergeben bekommen vom Type CAECommProjectApp.
	if (lpParam == NULL) {

		return 0;
	}

	// Referenz auf den uebergebenen Parameter
	CAECommProjectApp &own = *(CAECommProjectApp*)lpParam;

	// Hauptschleife ausfuehren?
	int run = FALSE;

	int rc;
	short channelID;

    SOCKADDR_IN sa;
	struct timeval seltime;
	fd_set readfds;

	CChannel4 *channelConnect = NULL;
	
	// Locales Channel-Objekt fuer Trace usw..
	channelID = -1;
	CChannel4 &channel = *own.GetChannel4(channelID);
	channel.SetBusy(TRUE);

	// Zustandsstatus fuer Listen - Socket
	own.m_listenThreadState = DOWN;

	// Timeout fuer Verbindungsanfragen. 
	//Achtung: Nicht auf 0 da sonst der Thread blockiert wrde!
	seltime.tv_sec  = 1;
	seltime.tv_usec = 0;

	// Sockethandle reservieren
    own.m_listenSocket = socket(AF_INET, SOCK_STREAM, 0);
    if(own.m_listenSocket != INVALID_SOCKET) {

		memset(&sa, 0, sizeof(sa));

		sa.sin_family = AF_INET;
		sa.sin_port = htons(own.m_listenPort);

		// Socket mit lokalem Port verbinden.
		if(bind(own.m_listenSocket, (SOCKADDR *)&sa, sizeof(sa)) == SOCKET_ERROR) {
        
			if(WSAGetLastError() == WSAEADDRINUSE) {

				channel.Trace("%s:(%i):ERROR(1): CAECommProjectApp::ListenThread [ bind fail (%i) Die gewhlte Portnummer ist in Gebrauch (%i) ]\n", __FILE__, __LINE__, WSAGetLastError(), own.m_listenPort);
			} else {

				channel.Trace("%s:(%i):ERROR(1): CAECommProjectApp::ListenThread [ bind fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
			}

			if (closesocket(own.m_listenSocket) == SOCKET_ERROR) {

				channel.Trace("%s:(%i):ERROR(1): CAECommProjectApp::ListenThread [ closesocket fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
			}
			own.m_listenSocket = INVALID_SOCKET;
		} else {

			// Mit Reset - Status beginnen
			run = TRUE;
			own.m_listenThreadState = RESET;
		}
	} else {

		channel.Trace("%s:(%i):ERROR(1): CAECommProjectApp::ListenThread [ socket fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
	}

	// Hauptschleife des Worker
	while (run) {

		// Beendigungssignal abfragen
		if (own.m_endSignal[0]) {

			// Listensocket herunterfahren
			if (own.m_listenSocket != INVALID_SOCKET) {

				if (own.m_listenThreadState != DOWN) {

					// Anstehende Verbindungsgesuche verwerfen
					if(listen(own.m_listenSocket, 1) == SOCKET_ERROR) { 

						channel.Trace("%s:(%i):ERROR(1): CAECommProjectApp::ListenThread [ listen fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
					}
				}
				// Listensocket schliessen
				if (closesocket(own.m_listenSocket) == SOCKET_ERROR) {

					channel.Trace("%s:(%i):ERROR(1): CAECommProjectApp::ListenThread [ closesocket fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
				}
				own.m_listenSocket = INVALID_SOCKET;

				// Eventuelle Verbindungen beenden
				for (int i=0; i < MAX_CHANNEL4; i++) {
					if ((own.m_channel4Array[i] != NULL) && (own.m_channel4Array[i]->IsBusy())) {

						own.m_channel4Array[i]->DirectClose();
					}
				}
			}
			
			// Zustand setzen
			own.m_listenThreadState = DOWN;

			// Hauptschleife verlassen
			run = FALSE;
			continue;
		}

		// Je nach Zustand den Listen-Socket verwalten
		switch(own.m_listenThreadState) {
		case DOWN: 
				// Aufgrund eines Fehlers abgeschaltet
			break;

		case RESET: // Verwirft eventuell anstehende Verbindungsanforderungen,
			        // durch (oder) Erzeugung Listensocket.
			if(listen(own.m_listenSocket, 1) == SOCKET_ERROR) { 

				channel.Trace("%s:(%i):ERROR(1): CAECommProjectApp::ListenThread [ listen fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());

				own.m_listenThreadState = DOWN;
				channel.Trace("%s:(%i):INFO(1): CAECommProjectApp::ListenThread [ State change to down... ]\n", __FILE__, __LINE__);
			} else {

				own.m_listenThreadState = LISTENING;
				channel.Trace("%s:(%i):INFO(1): CAECommProjectApp::ListenThread [ State change to listening... ]\n", __FILE__, __LINE__);
			}
			break;

		case LISTENING: // Wartet auf Connect. Im "listening" - Zustand.

			FD_ZERO((fd_set FAR *) &readfds);
			FD_SET(own.m_listenSocket, (fd_set FAR *) &readfds);

			if ((rc = select(FD_SETSIZE,
				(fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
				(const struct timeval FAR *)&seltime)) == SOCKET_ERROR) {

				channel.Trace("%s:(%i):ERROR(1): CAECommProjectApp::ListenThread [ select fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
			} else if (rc) {

				// Verbindungsanforderung auf Listensocket im naechsten
				// Druchlauf verarbeiten.
				own.m_listenThreadState = ACCEPTING;
				channel.Trace("%s:(%i):INFO(1): CAECommProjectApp::ListenThread [ State change to accepting... ]\n", __FILE__, __LINE__);
			} 
			break;

		case ACCEPTING: // Verbindungsanforderung verarbeiten

			channelID = -1;
			channelConnect = own.GetChannel4(channelID);

			channel.Trace("%s:(%i):INFO(1): CAECommProjectApp::ListenThread [ Try connect to (%i) ]\n", __FILE__, __LINE__, channelID);

			if ((channelConnect == NULL) || !channelConnect->Connect(&own.m_listenSocket, channelID, 254)) {

				// Wenn alle Kanalobjekte vergeben wurden
				// oder ein Fehler bei der Erstellung auftrat, dann
				// Verbindungsanforderung verwerfen.
				if (closesocket(accept(own.m_listenSocket, NULL, NULL)) == SOCKET_ERROR) {

					channel.Trace("%s:(%i):INFO(1): CAECommProjectApp::ListenThread [ closesocket fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
				}

				own.m_listenThreadState = RESET;
				channel.Trace("%s:(%i):INFO(1): CAECommProjectApp::ListenThread [ State change to reset... ]\n", __FILE__, __LINE__);
				continue;
			}

			// Verbindungsanfrage verarbeitet
			own.m_listenThreadState = LISTENING;
			channel.Trace("%s:(%i):INFO(1): CAECommProjectApp::ListenThread [ State change to listening... ]\n", __FILE__, __LINE__);
			break;
		}
	}

	// Lokales Objekt freigeben
	channel.SetBusy(FALSE);

	// Threadzeiger ruecksetzen
	own.m_workerThread[0] = NULL;

	// Thread Beendigung ruecksetzen
	own.m_endSignal[0] = FALSE;

	return 0;
}

unsigned long WINAPI CAECommProjectApp::NotifyThread(LPVOID lpParam)
{
	// Muss einen Parameter uebergeben bekommen vom Type CAECommProjectApp.
	if (lpParam == NULL) {

		return 0;
	}

	// Referenz auf den uebergebenen Parameter
	CAECommProjectApp &own = *(CAECommProjectApp*)lpParam;

	// Hauptschleife ausfuehren?
	int run = TRUE;

	// Hauptschleife des Worker
	while (run) {

		// Beendigungssignal abfragen
		if (own.m_endSignal[1]) {

			// Hauptschleife verlassen
			run = FALSE;
			continue;
		}

		int deviceType, serialNum, commType, commStringLength, commValue, deviceIDRemote, error;
		char commString[64];
		commStringLength = sizeof(commString);
		if (own.RecordAvailable(&deviceType, &serialNum, &commType, commString, &commStringLength, &commValue, &deviceIDRemote, &error) == IV_ERROR_SUCCESS) {

			// Use introspection API to look up the supported methods
			XmlRpc::XmlRpcClient c("192.168.123.177", 9000);
			XmlRpc::XmlRpcValue params, result;
			params[0] = deviceType;
			params[1] = serialNum;
			params[2] = commType;
			params[3] = commString;
			params[4] = commValue;
			params[5] = deviceIDRemote;

			if (c.execute("DFCRecordAvailable", params, result)) {
				
				TRACE("DFCRecordAvailable: %i\n", int(result));
			}
		}


		Sleep(500);
	}

	// Threadzeiger ruecksetzen
	own.m_workerThread[1] = NULL;

	// Thread Beendigung ruecksetzen
	own.m_endSignal[1] = FALSE;

	return 0;
}
#endif
/////////////////////////////////////////////////////////////////////////////
// Das einzige CAECommProjectApp-Objekt

CAECommProjectApp theApp;

